@@ -34,7 +34,10 @@ The library has special public value for nonexistent or expired keys called
     from oslo_cache import core
     NO_VALUE = core.NO_VALUE
 """
+import re
+import socket
 import ssl
+import urllib.parse
 
 import dogpile.cache
 from dogpile.cache import api
@@ -100,6 +103,18 @@ class _DebugProxy(proxy.ProxyBackend):
         self.proxied.delete_multi(keys)
 
 
+def _parse_sentinel(sentinel):
+    # IPv6 (eg. [::1]:6379 )
+    match = re.search(r'^\[(\S+)\]:(\d+)$', sentinel)
+    if match:
+        return (match[1], int(match[2]))
+    # IPv4 or hostname (eg. 127.0.0.1:6379 or localhost:6379)
+    match = re.search(r'^(\S+):(\d+)$', sentinel)
+    if match:
+        return (match[1], int(match[2]))
+    raise exception.ConfigurationError('Malformed sentinel server format')
+
+
 def _build_cache_config(conf):
     """Build the cache region dictionary configuration.
 
@@ -107,10 +122,10 @@ def _build_cache_config(conf):
     """
     prefix = conf.cache.config_prefix
     conf_dict = {}
-    conf_dict['%s.backend' % prefix] = _opts._DEFAULT_BACKEND
+    conf_dict[f'{prefix}.backend'] = _opts._DEFAULT_BACKEND
     if conf.cache.enabled is True:
-        conf_dict['%s.backend' % prefix] = conf.cache.backend
-    conf_dict['%s.expiration_time' % prefix] = conf.cache.expiration_time
+        conf_dict[f'{prefix}.backend'] = conf.cache.backend
+    conf_dict[f'{prefix}.expiration_time'] = conf.cache.expiration_time
     for argument in conf.cache.backend_argument:
         try:
             (argname, argvalue) = argument.split(':', 1)
@@ -136,81 +151,188 @@ def _build_cache_config(conf):
         conf_dict[arg_key] = argvalue
 
         _LOG.debug('Oslo Cache Config: %s', conf_dict)
-    # NOTE(yorik-sar): these arguments will be used for memcache-related
-    # backends. Use setdefault for url to support old-style setting through
-    # backend_argument=url:127.0.0.1:11211
-    #
-    # NOTE(morgan): If requested by config, 'flush_on_reconnect' will be set
-    # for pooled connections. This can ensure that stale data is never
-    # consumed from a server that pops in/out due to a network partition
-    # or disconnect.
-    #
-    # See the help from python-memcached:
-    #
-    # param flush_on_reconnect: optional flag which prevents a
-    #        scenario that can cause stale data to be read: If there's more
-    #        than one memcached server and the connection to one is
-    #        interrupted, keys that mapped to that server will get
-    #        reassigned to another. If the first server comes back, those
-    #        keys will map to it again. If it still has its data, get()s
-    #        can read stale data that was overwritten on another
-    #        server. This flag is off by default for backwards
-    #        compatibility.
-    #
-    # The normal non-pooled clients connect explicitly on each use and
-    # does not need the explicit flush_on_reconnect
-    conf_dict.setdefault('%s.arguments.url' % prefix,
-                         conf.cache.memcache_servers)
-    for arg in ('dead_retry', 'socket_timeout', 'pool_maxsize',
-                'pool_unused_timeout', 'pool_connection_get_timeout',
-                'pool_flush_on_reconnect', 'sasl_enabled', 'username',
-                'password'):
-        value = getattr(conf.cache, 'memcache_' + arg)
-        conf_dict['%s.arguments.%s' % (prefix, arg)] = value
+
+    if conf.cache.backend == 'dogpile.cache.redis':
+        if conf.cache.redis_password is None:
+            netloc = conf.cache.redis_server
+        else:
+            if conf.cache.redis_username:
+                netloc = '%s:%s@%s' % (conf.cache.redis_username,
+                                       conf.cache.redis_password,
+                                       conf.cache.redis_server)
+            else:
+                netloc = ':%s@%s' % (conf.cache.redis_password,
+                                     conf.cache.redis_server)
+
+        parts = urllib.parse.ParseResult(
+            scheme=('rediss' if conf.cache.tls_enabled else 'redis'),
+            netloc=netloc, path=str(conf.cache.redis_db), params='', query='',
+            fragment='')
+
+        conf_dict.setdefault(
+            f'{prefix}.arguments.url',
+            urllib.parse.urlunparse(parts)
+        )
+        for arg in ('socket_timeout',):
+            value = getattr(conf.cache, 'redis_' + arg)
+            conf_dict[f'{prefix}.arguments.{arg}'] = value
+    elif conf.cache.backend == 'dogpile.cache.redis_sentinel':
+        for arg in ('username', 'password', 'socket_timeout', 'db'):
+            value = getattr(conf.cache, 'redis_' + arg)
+            conf_dict[f'{prefix}.arguments.{arg}'] = value
+        conf_dict[f'{prefix}.arguments.service_name'] = \
+            conf.cache.redis_sentinel_service_name
+        if conf.cache.redis_sentinels:
+            conf_dict[f'{prefix}.arguments.sentinels'] = [
+                _parse_sentinel(s) for s in conf.cache.redis_sentinels]
+    else:
+        # NOTE(yorik-sar): these arguments will be used for memcache-related
+        # backends. Use setdefault for url to support old-style setting through
+        # backend_argument=url:127.0.0.1:11211
+        #
+        # NOTE(morgan): If requested by config, 'flush_on_reconnect' will be
+        # set for pooled connections. This can ensure that stale data is never
+        # consumed from a server that pops in/out due to a network partition
+        # or disconnect.
+        #
+        # See the help from python-memcached:
+        #
+        # param flush_on_reconnect: optional flag which prevents a
+        #        scenario that can cause stale data to be read: If there's more
+        #        than one memcached server and the connection to one is
+        #        interrupted, keys that mapped to that server will get
+        #        reassigned to another. If the first server comes back, those
+        #        keys will map to it again. If it still has its data, get()s
+        #        can read stale data that was overwritten on another
+        #        server. This flag is off by default for backwards
+        #        compatibility.
+        #
+        # The normal non-pooled clients connect explicitly on each use and
+        # does not need the explicit flush_on_reconnect
+        conf_dict.setdefault(f'{prefix}.arguments.url',
+                             conf.cache.memcache_servers)
+
+        for arg in ('dead_retry', 'socket_timeout', 'pool_maxsize',
+                    'pool_unused_timeout', 'pool_connection_get_timeout',
+                    'pool_flush_on_reconnect', 'sasl_enabled', 'username',
+                    'password'):
+            value = getattr(conf.cache, 'memcache_' + arg)
+            conf_dict[f'{prefix}.arguments.{arg}'] = value
 
     if conf.cache.tls_enabled:
-        _LOG.debug('Oslo Cache TLS - CA: %s', conf.cache.tls_cafile)
-        tls_context = ssl.create_default_context(cafile=conf.cache.tls_cafile)
+        if conf.cache.backend in ('dogpile.cache.bmemcache',
+                                  'dogpile.cache.pymemcache',
+                                  'oslo_cache.memcache_pool'):
+            _LOG.debug('Oslo Cache TLS - CA: %s', conf.cache.tls_cafile)
+            tls_context = ssl.create_default_context(
+                cafile=conf.cache.tls_cafile)
 
-        if conf.cache.tls_certfile is not None:
-            _LOG.debug('Oslo Cache TLS - cert: %s', conf.cache.tls_certfile)
-            _LOG.debug('Oslo Cache TLS - key: %s', conf.cache.tls_keyfile)
-            tls_context.load_cert_chain(
-                conf.cache.tls_certfile,
-                conf.cache.tls_keyfile,
+            if conf.cache.enforce_fips_mode:
+                if hasattr(ssl, 'FIPS_mode'):
+                    _LOG.info("Enforcing the use of the OpenSSL FIPS mode")
+                    ssl.FIPS_mode_set(1)
+                else:
+                    raise exception.ConfigurationError(
+                        "OpenSSL FIPS mode is not supported by your Python "
+                        "version. You must either change the Python "
+                        "executable used to a version with FIPS mode support "
+                        "or disable FIPS mode by setting "
+                        "the '[cache] enforce_fips_mode' configuration option "
+                        "to 'False'.")
+
+            if conf.cache.tls_certfile is not None:
+                _LOG.debug('Oslo Cache TLS - cert: %s',
+                           conf.cache.tls_certfile)
+                _LOG.debug('Oslo Cache TLS - key: %s', conf.cache.tls_keyfile)
+                tls_context.load_cert_chain(
+                    conf.cache.tls_certfile,
+                    conf.cache.tls_keyfile,
+                )
+
+            if conf.cache.tls_allowed_ciphers is not None:
+                _LOG.debug(
+                    'Oslo Cache TLS - ciphers: %s',
+                    conf.cache.tls_allowed_ciphers,
+                )
+                tls_context.set_ciphers(conf.cache.tls_allowed_ciphers)
+
+            conf_dict[f'{prefix}.arguments.tls_context'] = tls_context
+        elif conf.cache.backend in ('dogpile.cache.redis',
+                                    'dogpile.cache.redis_sentinel'):
+            if conf.cache.tls_allowed_ciphers is not None:
+                raise exception.ConfigurationError(
+                    "Limiting allowed ciphers is not supported by "
+                    "the %s backend" % conf.cache.backend)
+            if conf.cache.enforce_fips_mode:
+                raise exception.ConfigurationError(
+                    "FIPS mode is not supported by the %s backend" %
+                    conf.cache.backend)
+
+            conn_kwargs = {}
+            if conf.cache.tls_cafile is not None:
+                _LOG.debug('Oslo Cache TLS - CA: %s', conf.cache.tls_cafile)
+                conn_kwargs['ssl_ca_certs'] = conf.cache.tls_cafile
+            if conf.cache.tls_certfile is not None:
+                _LOG.debug('Oslo Cache TLS - cert: %s',
+                           conf.cache.tls_certfile)
+                _LOG.debug('Oslo Cache TLS - key: %s', conf.cache.tls_keyfile)
+                conn_kwargs.update({
+                    'ssl_certfile': conf.cache.tls_certfile,
+                    'ssl_keyfile': conf.cache.tls_keyfile
+                })
+            if conf.cache.backend == 'dogpile.cache.redis_sentinel':
+                conn_kwargs.update({'ssl': True})
+                conf_dict[f'{prefix}.arguments.connection_kwargs'] = \
+                    conn_kwargs
+                conf_dict[f'{prefix}.arguments.sentinel_kwargs'] = \
+                    conn_kwargs
+            else:
+                conf_dict[f'{prefix}.arguments.connection_kwargs'] = \
+                    conn_kwargs
+        else:
+            raise exception.ConfigurationError(
+                "TLS setting via [cache] tls_enabled is not supported by the "
+                "%s backend. Set [cache] tls_enabled=False or use a different "
+                "backend." % conf.cache.backend
             )
 
-        if conf.cache.tls_allowed_ciphers is not None:
-            _LOG.debug(
-                'Oslo Cache TLS - ciphers: %s',
-                conf.cache.tls_allowed_ciphers,
-            )
-            tls_context.set_ciphers(conf.cache.tls_allowed_ciphers)
-
-        conf_dict['%s.arguments.tls_context' % prefix] = tls_context
-
-    # NOTE(hberaud): Pymemcache support socket keepalive, If it is enable in
-    # our config then configure it to enable this feature.
-    # The socket keepalive feature means that pymemcache will be able to check
+    # NOTE(hberaud): Pymemcache backend and redis backends support socket
+    # keepalive, If it is enable in our config then configure it to enable this
+    # feature.
+    # The socket keepalive feature means that client will be able to check
     # your connected socket and determine whether the connection is still up
     # and running or if it has broken.
     # This could be used by users who want to handle fine grained failures.
     if conf.cache.enable_socket_keepalive:
-        if conf.cache.backend != 'dogpile.cache.pymemcache':
-            msg = _(
-                "Socket keepalive is only supported by the "
-                "'dogpile.cache.pymemcache' backend."
+        if conf.cache.backend == 'dogpile.cache.pymemcache':
+            import pymemcache
+            socket_keepalive = pymemcache.KeepaliveOpts(
+                idle=conf.cache.socket_keepalive_idle,
+                intvl=conf.cache.socket_keepalive_interval,
+                cnt=conf.cache.socket_keepalive_count)
+            # As with the TLS context above, the config dict below will be
+            # consumed by dogpile.cache that will be used as a proxy between
+            # oslo.cache and pymemcache.
+            conf_dict[f'{prefix}.arguments.socket_keepalive'] = \
+                socket_keepalive
+        elif conf.cache.backend in ('dogpile.cache.redis',
+                                    'dogpile.cache.redis_sentinel'):
+            socket_keepalive_options = {
+                socket.TCP_KEEPIDLE: conf.cache.socket_keepalive_idle,
+                socket.TCP_KEEPINTVL: conf.cache.socket_keepalive_interval,
+                socket.TCP_KEEPCNT: conf.cache.socket_keepalive_count
+            }
+            conf_dict.setdefault(
+                f'{prefix}.arguments.connection_kwargs', {}
+            ).update({
+                'socket_keepalive': True,
+                'socket_keepalive_options': socket_keepalive_options
+            })
+        else:
+            raise exception.ConfigurationError(
+                "Socket keepalive is not supported by the %s backend"
+                % conf.cache.backend
             )
-            raise exception.ConfigurationError(msg)
-        import pymemcache
-        socket_keepalive = pymemcache.KeepaliveOpts(
-            idle=conf.cache.socket_keepalive_idle,
-            intvl=conf.cache.socket_keepalive_interval,
-            cnt=conf.cache.socket_keepalive_count)
-        # As with the TLS context above, the config dict below will be
-        # consumed by dogpile.cache that will be used as a proxy between
-        # oslo.cache and pymemcache.
-        conf_dict['%s.arguments.socket_keepalive' % prefix] = socket_keepalive
 
     # NOTE(hberaud): The pymemcache library comes with retry mechanisms that
     # can be used to wrap all kind of pymemcache clients. The retry wrapper
@@ -225,16 +347,16 @@ def _build_cache_config(conf):
             )
             raise exception.ConfigurationError(msg)
         import pymemcache
-        conf_dict['%s.arguments.enable_retry_client' % prefix] = True
-        conf_dict['%s.arguments.retry_attempts' % prefix] = \
+        conf_dict[f'{prefix}.arguments.enable_retry_client'] = True
+        conf_dict[f'{prefix}.arguments.retry_attempts'] = \
             conf.cache.retry_attempts
-        conf_dict['%s.arguments.retry_delay' % prefix] = \
+        conf_dict[f'{prefix}.arguments.retry_delay'] = \
             conf.cache.retry_delay
-        conf_dict['%s.arguments.hashclient_retry_attempts' % prefix] = \
+        conf_dict[f'{prefix}.arguments.hashclient_retry_attempts'] = \
             conf.cache.hashclient_retry_attempts
-        conf_dict['%s.arguments.hashclient_retry_delay' % prefix] = \
+        conf_dict[f'{prefix}.arguments.hashclient_retry_delay'] = \
             conf.cache.hashclient_retry_delay
-        conf_dict['%s.arguments.dead_timeout' % prefix] = \
+        conf_dict[f'{prefix}.arguments.dead_timeout'] = \
             conf.cache.dead_timeout
 
     return conf_dict
@@ -324,7 +446,7 @@ def configure_cache_region(conf, region):
 
         config_dict = _build_cache_config(conf)
         region.configure_from_config(config_dict,
-                                     '%s.' % conf.cache.config_prefix)
+                                     f'{conf.cache.config_prefix}.')
 
         if conf.cache.debug_cache_backend:
             region.wrap(_DebugProxy)
